import yaml
from collections import defaultdict
from quart import Blueprint, request, g
from quart.json import jsonify
from sovereign import XDS_TEMPLATES
from sovereign.utils.mock import mock_discovery_request
from sovereign.utils.templates import remove_tls_certificates
from sovereign import discovery
from sovereign.sources import load_sources
from sovereign.decorators import cache

blueprint = Blueprint('admin', __name__)

latest_version = list(XDS_TEMPLATES)[-1]
discovery_types = XDS_TEMPLATES[latest_version].keys()


@blueprint.route('/admin/xds_dump')
def display_config():
    xds_type = request.args.get('type')
    service_cluster = request.args.get('partition', '')
    resource_names = request.args.get('resource_names')
    region = request.args.get('region')
    version = request.args.get('envoy_version', '1.9.0')
    ret = defaultdict(list)
    code = 200

    if xds_type in discovery_types:
        selected_types = [xds_type]
    else:
        selected_types = []
        ret = {
            'message': 'Query parameter "type" must be one of ["clusters", "listeners", "routes", "endpoints"]'
        }
        code = 400

    for discovery_type in selected_types:
        mock_request = mock_discovery_request(
            service_cluster=service_cluster,
            resource_names=resource_names,
            region=region
        )
        response = discovery.response(
            request=mock_request,
            xds=discovery_type,
            version=version,
            debug=True
        )
        if isinstance(response, dict):
            ret['resources'] += response.get('resources') or []

    # Hide private keys
    for resource in ret.get('resources', []):
        if not isinstance(resource, dict):
            continue
        if 'Listener' not in resource['@type']:
            continue
        remove_tls_certificates(resource)

    return jsonify(ret), code


@blueprint.route('/admin/source_dump')
def instances():
    debug = yaml.safe_load(request.args.get('debug', 'no'))
    cluster = request.args.get('partition', '')
    modified = yaml.safe_load(request.args.get('modified', 'yes'))
    args = {
        'debug': debug,
        'modify': modified,
        'service_cluster': cluster
    }
    ret = load_sources(**args)
    g.log = g.log.bind(args=args)
    return jsonify(ret)


@blueprint.route('/admin/cache_dump')
def show_cached_keys():
    return jsonify(list(sorted(cache._cache.keys())))


@blueprint.route('/admin/cache_purge')
def purge_cache():
    cache.clear()
    return 'Cache cleared'



"""
File Source
-----------

Example(s) configuration file (YAML):

.. code-block:: yaml

   sources:
     - type: file
       config:
         path: file:///tmp/instances.yaml
     - type: file
       config:
         path: file+json:///tmp/instances.json
     - type: file
       config:
         path: https://mywebsite.com/configuration/instances.yaml
     - type: file
       config:
         path: https+json://mywebsite.com/configuration/instances.json
     - type: file
       config:
         path: https+json://mywebsite.com/configuration/instances.json
     - type: file
       config:
         path: pkgdata://sovereign:config/instances.yaml

Example of what the contents should look like:

.. code-block:: yaml

   - instance_id: <identifier>
     service_clusters:
       - P2
     parameters:
       clusters:
         - name: upstream
           healthchecks:
             - path: /healthcheck
           hosts:
             - address: aws.amazon.com
               port: 443
               region: us-east-1
         vhosts:
         - clusters:
             - name: upstream
           domains:
             - aws.amazon.com
             - amazon.dev.globaledge.internal
           name: vhost
           rewrite: 'yes'
"""
from sovereign.sources.lib import Source
from sovereign.config_loader import load
from sovereign.decorators import memoize


class File(Source):
    def __init__(self, *args, **kwargs):
        super(File, self).__init__(*args, **kwargs)
        for arg in args:
            try:
                self.path = arg['path']
                break
            except KeyError:
                pass
        else:
            raise KeyError('File source needs to specify "path" within config')

    def get(self):
        """
        Uses the file config loader to load the given path
        """
        return self._file_source_get(self.path)

    @staticmethod
    @memoize(30)
    def _file_source_get(path):
        return load(path)



import os
from collections import defaultdict
from datadog import statsd
from sovereign import config_loader
from sovereign.logs import LOG
try:
    import sentry_sdk
    from sentry_sdk.integrations.flask import FlaskIntegration
except ImportError:
    sentry_sdk = None


XDS_TEMPLATES = defaultdict(dict)
TEMPLATE_CONTEXT = dict()
DEBUG = bool(os.getenv('SOVEREIGN_DEBUG'))
ENVIRONMENT = os.getenv('SOVEREIGN_ENVIRONMENT_TYPE', os.getenv('MICROS_ENVTYPE', 'local'))
CONFIG = dict()

try:
    CONFIG_PATHS = os.getenv('SOVEREIGN_CONFIG').split(',')
except AttributeError:
    LOG.error('No configuration specified via environment variable SOVEREIGN_CONFIG')
else:
    for path in CONFIG_PATHS:
        cfg = config_loader.load(path)
        CONFIG.update(cfg)

    _templates = CONFIG['templates'].items()
    _template_context = CONFIG.get('template_context', {}).items()

    for version, templates in _templates:
        for _type, path in templates.items():
            XDS_TEMPLATES[version][_type] = config_loader.load(path)

    for key, value in _template_context:
        TEMPLATE_CONTEXT[key] = config_loader.load(value)

    if CONFIG.get('statsd', {}).get('enabled'):
        statsd.host = CONFIG['statsd']['host']
        statsd.namespace = CONFIG['statsd'].get('namespace', 'sovereign')
        for tag, value in CONFIG['statsd'].get('tags', {}).items():
            value = config_loader.load(value)
            statsd.constant_tags.extend([f'{tag}:{value}'])

    NO_CHANGE_CODE = CONFIG.get('no_changes_response_code', 304)

    SENTRY_DSN = os.getenv('SOVEREIGN_SENTRY_DSN')
    if all([SENTRY_DSN, sentry_sdk]):
        sentry_sdk.init(
            SENTRY_DSN,
            integrations=[
                FlaskIntegration()
            ]
        )

    LOG.msg(
        event='startup',
        envtype=os.getenv('MICROS_ENVTYPE', os.getenv('SOVEREIGN_ENVIRONMENT_TYPE')),
        env=ENVIRONMENT,
        config=CONFIG_PATHS,
        context=_template_context,
        templates=_templates,
        is_debug=DEBUG
    )



"""
Sources are used to retrieve arbitrary data that can then be represented
in templates as envoy configuration.

A new custom source can be added by subclassing the Source class and
adding an entry point using setuptools.

`todo add guide for adding entry points`

When the control plane receives a request for configuration from an envoy
proxy, it first runs through all the configured sources and combines the
data received from them into a single list, to be used within templates.

The results are cached for 30 seconds to allow several proxies to poll
at the same time, receiving data that is consistent with each other.
`todo, make the cache expiry configurable`
"""
from typing import List
from pkg_resources import iter_entry_points
from sovereign import CONFIG, DEBUG, statsd
from sovereign.logs import LOG
from sovereign.modifiers import apply_modifications

_entry_points = iter_entry_points('sovereign.sources')
_sources = {e.name: e.load() for e in _entry_points}

if not _sources:
    raise RuntimeError('No sources available.')


def load_sources(service_cluster, modify=True, debug=DEBUG) -> List[dict]:
    """
    Runs all configured Sources, returning a list of the combined results

    :param service_cluster: filter results by envoy service cluster `todo` make filtering configurable
    :param modify: switch to enable or disable modifications via Modifiers
    :param debug: switch mainly used for testing
    :return: a list of dictionaries
    """
    ret = list()
    for source in _enumerate_sources():
        if not isinstance(source, dict):
            LOG.msg('Received a non-dictionary source', level='warn', source_repr=repr(source))
            continue

        # TODO: create a 'node match' configurable option
        #       which maps between source/instance <-> node (aka envoy node)
        envoy_service_clusters = source.get('service_clusters', [])
        conditions = (
            service_cluster in envoy_service_clusters,
            envoy_service_clusters == ['*'],
            service_cluster == '' and debug
        )
        if any(conditions):
            ret.append(source)
    if modify:
        return apply_modifications(ret)
    return ret


def _enumerate_sources():
    for source in CONFIG['sources']:
        name = source['type']
        config = source['config']
        yield from _enumerate_source(name, config)


def _enumerate_source(name, conf):
    with statsd.timed('sources.load_ms', tags=[f'source_name:{name}'], use_ms=True):
        source = _sources[name]
        _source = source(conf)
        _source.setup()
        yield from _source.get()



def mock_discovery_request(service_cluster=None, resource_names=None, region='none'):
    return {
        'node': {
            # A blank string will select any cluster when debugging
            'cluster': service_cluster or '',
            'build_version': 'e5f864a82d4f27110359daa2fbdcb12d99e415b9/1.8.0/Clean/RELEASE',
            "locality": {
                "zone": region
            },
        },
        'version_info': '0',
        'resource_names': resource_names or []
    }



import gzip
from functools import wraps
from datetime import timedelta
from quart import after_this_request, request, Response
from quart.datastructures import HeaderSet
from werkzeug.exceptions import BadRequest, Unauthorized
from cachelib import SimpleCache
from sovereign import CONFIG, statsd
from sovereign.utils.crypto import decrypt, KEY_AVAILABLE, InvalidToken


cache = SimpleCache()


def memoize(timeout):
    """
    Decorator to cache a function by name/args

    :param timeout: How long to keep the result
    """
    if isinstance(timeout, timedelta):
        timeout = timeout.seconds

    def decorator(decorated):
        @wraps(decorated)
        def wrapper(*args, **kwargs):
            key = f'{decorated.__name__}{args}{kwargs}'
            ret = cache.get(key)
            metrics_tags = [
                f'function:{decorated.__name__}'
            ]
            if ret is None:
                statsd.increment('cache.miss', tags=metrics_tags)
                ret = decorated(*args, **kwargs)
                cache.set(key, ret, timeout=timeout)
            else:
                statsd.increment('cache.hit', tags=metrics_tags)
            return ret
        return wrapper
    return decorator


def envoy_authorization_required(decorated):
    """
    Decorator that checks a functions args for something that resembles
    an envoy discovery request, and attempts to decrypt the authorization that
    it contains.

    Raises an exception if there is no discovery request, or if the
    authorization token fails to decrypt.
    """
    @wraps(decorated)
    def wrapper(*args, **kwargs):
        auth_enabled_ = CONFIG.get('auth_required') or CONFIG.get('auth_enabled')
        if auth_enabled_ and not KEY_AVAILABLE:
            raise RuntimeError('No Fernet key loaded, and auth is enabled.')
        if not kwargs.get('debug') and auth_enabled_:
            for arg in args:
                if _request_contains_valid_auth(arg):
                    statsd.increment('discovery.auth.success')
                    break
            else:
                # No arg could be found with auth
                statsd.increment('discovery.auth.failed')
                raise Unauthorized('No authentication provided')
        return decorated(*args, **kwargs)
    return wrapper


def _request_contains_valid_auth(wrapped_fn_argument):
    try:
        metadata = wrapped_fn_argument['node']['metadata']
        auth = metadata.pop('auth')  # Consume the auth, it's not needed past here
        decrypt(auth)
        return True
    except (KeyError, AttributeError):
        pass
    except InvalidToken:
        raise Unauthorized('The authentication provided was invalid')
    except Exception:
        raise BadRequest('The request was malformed')


def gzcompress(level=2, valid_codes=range(200, 304)):
    """
    Gzip compression decorator from : http://flask.pocoo.org/snippets/122/

    Modified to work with Quart
    Also added configurable compression level & status codes

    Unfortunately does not work with Envoy since it doesn't send gzip related headers
    """
    def decorator(f):
        @wraps(f)
        async def wrapper(*args, **kwargs):
            @after_this_request
            async def compress(response: Response) -> Response:
                accept_encoding = request.headers.get('Accept-Encoding', '').lower()

                if 'gzip' not in accept_encoding:
                    return response

                if response.status_code not in valid_codes or response.content_encoding:
                    return response

                data = await response.get_data()
                output = gzip.compress(data, compresslevel=level)

                response.set_data(output)
                response.content_encoding = 'gzip'
                response.vary = HeaderSet.from_header('Accept-Encoding')
                return response

            return await f(*args, **kwargs)
        return wrapper
    return decorator



from quart import Blueprint, request, jsonify, g
from sovereign import discovery, statsd, NO_CHANGE_CODE

blueprint = Blueprint('discovery', __name__)


@blueprint.route('/v2/discovery:<xds_type>', methods=['POST'])
async def discovery_endpoint(xds_type):
    discovery_request = await request.get_json(force=True)

    build_version = discovery_request['node']['build_version']
    revision, version, *other_metadata = build_version.split('/')
    resource_names = discovery_request.get('resource_names', [])

    g.log = g.log.bind(
        resource_names=resource_names,
        envoy_ver=version
    )

    response = discovery.response(discovery_request, xds_type, version)

    if not response['resources']:
        ret = 'No resources found'
        code = 404
    elif response['version_info'] == discovery_request.get('version_info'):
        ret = 'No changes'
        code = NO_CHANGE_CODE
    elif response['version_info'] != discovery_request.get('version_info', '0'):
        ret = response
        code = 200
    else:
        ret = 'Unknown Error'
        code = 500

    try:
        client_ip = discovery_request['node']['metadata']['ipv4']
    except KeyError:
        client_ip = '-'

    metrics_tags = [
        f"client_ip:{client_ip}",
        f"client_version:{version}",
        f"response_code:{code}",
        f"xds_type:{xds_type}"
    ]
    metrics_tags += [f"resource:{resource}" for resource in resource_names]
    statsd.increment('discovery.rq_total', tags=metrics_tags)
    return jsonify(ret), code



"""
Discovery
---------

Functions used to render and return discovery responses to Envoy proxies.

The templates are configurable. `todo See ref:Configuration#Templates`
"""
import hashlib
import yaml
from yaml.parser import ParserError
from sovereign import XDS_TEMPLATES, TEMPLATE_CONTEXT, DEBUG, statsd
from sovereign.decorators import envoy_authorization_required
from sovereign.sources import load_sources


def version_hash(config: dict) -> str:
    """
    Creates a 'version hash' to be used in envoy Discovery Responses.

    :param config: Any dictionary
    :return: 16 character hexadecimal string
    """
    config_string: bytes = repr(yaml.dump(config)).encode()
    return hashlib.sha256(config_string).hexdigest()


@envoy_authorization_required
def response(request, xds, version, debug=DEBUG) -> dict:
    """
    A Discovery **Request** typically looks something like:

    .. code-block:: json

        {
            "version_info": "0",
            "node": {
                "cluster": "T1",
                "build_version": "<revision hash>/<version>/Clean/RELEASE",
                "metadata": {
                    "auth": "..."
                }
            }
        }

    When we receive this, we give the client the latest configuration via a
    Discovery **Response** that looks something like this:

    .. code-block:: json

        {
            "version_info": "abcdef1234567890",
            "resources": []
        }

    The version_info is derived from :func:`sovereign.discovery.version_hash`

    :param request: An envoy Discovery Request
    :param xds: what type of XDS template to use when rendering
    :param version: what template version to render for (i.e. envoy 1.7.0, 1.8.0)
    :param debug: switch to control instance loading / exception raising
    :return: An envoy Discovery Response
    """
    partition = request['node']['cluster']
    context = {
        'instances': load_sources(partition, debug=debug),
        'resource_names': request.get('resource_names', []),
        'discovery_request': request,
        'debug': debug,
        **TEMPLATE_CONTEXT
    }
    metrics_tags = [
        f'xds_type:{xds}',
        f'partition:{partition}'
    ]
    if version not in XDS_TEMPLATES:
        version = 'default'
    template = XDS_TEMPLATES[version][xds]
    with statsd.timed('discovery.render_ms', use_ms=True, tags=metrics_tags):
        rendered = template.render(**context)
    try:
        configuration = yaml.load(rendered)
        configuration['version_info'] = version_hash(configuration)
        return configuration
    except ParserError:
        if debug:
            raise
        raise ParserError('Failed to render configuration')



"""
Service Broker source
---------------------

Example configuration (YAML):

.. code-block:: yaml

   sources:
     - type: service_broker
       config:
         # List of brokers
         brokers:
           - https://broker1.com/instances
           - https://broker2.net:8443/api/instances

         # Optional: somewhere to keep the last good config
         file: /tmp/broker_result_backup.json

         # Load from debug_instances when requests to brokers fail
         debug: yes
         debug_instances:
           - instance_id: my_service
             parameters:
               upstream_address: service.domain.com
             plan_id: 7d57270a-0348-58d3-829d-447a98fe98d5
             service_id: 10e5a402-45df-5afd-ae86-11377ce2bbb2
             service_clusters:
               - P2
"""
import json
import requests
from requests.exceptions import RequestException
from sovereign import DEBUG
from sovereign.decorators import memoize
from sovereign.sources.lib import Source

USER_AGENT = {
    'User-Agent': 'Envoy-Control-Plane (python-requests/{0})'.format(requests.__version__)
}


class ServiceBroker(Source):
    def __init__(self, *args, **kwargs):
        super(ServiceBroker, self).__init__(*args, **kwargs)
        for arg in args:
            if not isinstance(arg, dict):
                continue
            self.debug = arg.get('debug', DEBUG)
            self.debug_instances = arg.get('debug_instances', [])
            self.file = arg.get('file', './service_broker_instances_backup.json')
            self.brokers = arg.get('brokers', ['http://localhost:5000/v2/service_instances'])

    @staticmethod
    @memoize(30)
    def _get_from_broker(url):
        """
        Cached method that calls a GET to the broker

        :param url: one of the configured broker urls
        :return: HTTP response from the broker
        """
        return requests.get(url, headers=USER_AGENT, timeout=3)

    def get(self) -> list:
        """
        Retrieves data from the broker over http/s

        Returns a last known good configuration in the case of a failure.

        Returns debugging instances given via the service broker source configuration if
        debugging is enabled.

        :return: list of instances from the broker
        """
        request_failed = False
        for url in self.brokers:
            try:
                response = self._get_from_broker(url)
                instance_data = response.json()
            except RequestException:
                request_failed = True
            else:
                self.save(instance_data)
                return instance_data
        if request_failed and self.debug:
            return self.debug_instances
        return self.load()

    def save(self, data):
        """
        Saves a backup of the last known good configuration
        """
        with open(self.file, 'w+') as f:
            json.dump(data, f)

    def load(self):
        """
        Loads the last known good configuration in the case that
        the broker can't be contacted
        """
        with open(self.file) as f:
            return json.load(f)



import os
from collections import namedtuple
from cryptography.fernet import Fernet, InvalidToken

_legacy_key = os.getenv('FERNET_ENCRYPTION_KEY')
_secret_key = os.getenv('SOVEREIGN_ENCRYPTION_KEY', _legacy_key)


try:
    _cipher_suite = Fernet(_secret_key)
    KEY_AVAILABLE = True
except TypeError:
    KEY_AVAILABLE = False
    _cipher_suite = namedtuple('DisabledSuite', ['encrypt', 'decrypt'])
    _cipher_suite.encrypt = lambda x: b'Unavailable (No Secret Key)'
    _cipher_suite.decrypt = lambda x: b'Unavailable (No Secret Key)'


def encrypt(data: str, key=None) -> str:
    _local_cipher_suite = _cipher_suite
    if key is not None:
        _local_cipher_suite = Fernet(key)
    try:
        encrypted: bytes = _local_cipher_suite.encrypt(data.encode())
    except InvalidToken:
        raise InvalidToken('Encryption failed')
    except AttributeError:
        raise AttributeError('Encryption failed')
    return encrypted.decode()


def decrypt(data: str, key=None) -> str:
    _local_cipher_suite = _cipher_suite
    if key is not None:
        _local_cipher_suite = Fernet(key)
    try:
        decrypted: bytes = _local_cipher_suite.decrypt(data.encode())
    except InvalidToken:
        raise InvalidToken('Decryption failed')
    except AttributeError:
        raise AttributeError('Decryption failed')
    return decrypted.decode()


def generate_key() -> str:
    secret: bytes = Fernet.generate_key()
    return secret.decode()
